package com.winhc.bigdata.flink.jobs

import com.winhc.bigdata.flink.event.{ElasticSearchInfo, HbaseInfo, HologresInfo, UpdateEntity}
import com.winhc.bigdata.flink.func.{HbaseAsyncFunction, HbaseSinkFunction}
import com.winhc.bigdata.flink.implicits._
import com.winhc.bigdata.flink.java.constant.EnvConst
import com.winhc.bigdata.flink.java.entity.{PutCollection, SealedPut}
import com.winhc.bigdata.flink.jobs.company_job_step._
import com.winhc.bigdata.flink.source.KafkaSourceBuilder
import com.winhc.bigdata.flink.streaming.trigger.ProcessingTimeCountTrigger
import org.apache.commons.lang3.StringUtils
import org.apache.flink.api.common.eventtime.WatermarkStrategy
import org.apache.flink.api.common.functions.AggregateFunction
import org.apache.flink.api.scala._
import org.apache.flink.connector.kafka.source.KafkaSource
import org.apache.flink.streaming.api.functions.ProcessFunction
import org.apache.flink.streaming.api.scala.{AsyncDataStream, DataStream, StreamExecutionEnvironment}
import org.apache.flink.streaming.api.windowing.assigners.TumblingProcessingTimeWindows
import org.apache.flink.streaming.api.windowing.time.Time
import org.apache.flink.streaming.api.windowing.windows.TimeWindow
import org.apache.flink.util.Collector
import org.apache.hadoop.hbase.client.Put
import org.json4s.JsonAST.{JNull, JObject, JValue}
import org.json4s.{JNothing, JString}
import java.nio.charset.StandardCharsets
import java.util.concurrent.TimeUnit
import com.alibaba.dcm.DnsCacheManipulator
import scala.collection.immutable


/**
 * @author ZhangJi
 * @since 2021-10-14 10:11
 */
object TestJob1 {
  def main(args: Array[String]): Unit = {
    DnsCacheManipulator.loadDnsCacheConfig()

    val env = StreamExecutionEnvironment.getExecutionEnvironment
    //env.getCheckpointConfig.setCheckpointTimeout(60000*5)

    env.getConfig.setGlobalJobParameters(EnvConst.createParameterTool(args))
    val kafkaSource: KafkaSource[String] = KafkaSourceBuilder.buildSourceFunction("flink_test")

    val source = env.fromSource(kafkaSource, WatermarkStrategy.forMonotonousTimestamps(), "Kafka Source")

    val allSource: DataStream[UpdateEntity] = source.transform_and_validation().name("transform_and_validation")
    allSource.getSideOutput(OutputTags.ExceptionTag).map(e => e.toJson()).print()
    allSource.getSideOutput(OutputTags.TransformErrorTag).map(e => e.toJson()).print()
    allSource.getSideOutput(OutputTags.ValidationErrorTag).map(e => e.toJson()).print()
    //val asyncDataStream = AsyncDataStream.unorderedWait(allSource, new HbaseAsyncFunction, 10, TimeUnit.SECONDS)

    val afterOutput: DataStream[(String, Map[String, JObject])] = allSource.process(new ProcessFunction[UpdateEntity, (String, Map[String, JObject])] {
      override def processElement(ue: UpdateEntity, ctx: ProcessFunction[UpdateEntity, (String, Map[String, JObject])]#Context, out: Collector[(String, Map[String, JObject])]): Unit = {
        val company = ue.company
        if (company != null) {
          ctx.output(OutputTags.HBASE_SINK_TAG, ("NG_RT_COMPANY", "F", ue.companyId(), company))
          ctx.output(OutputTags.HOLO_SINK_TAG, ("ng_rt_company", "public", ue.companyId(), company))
        }
        for (table <- ue.dims.values) {
          val meta = table.metaInfo
          val data = table.data
          meta.hbase match {
            case Some(hi) =>
              val HbaseInfo(h0, h1) = hi
              data.foreach(t => {
                val (rowkey, jo) = t
                ctx.output(OutputTags.HBASE_SINK_TAG, (h0, h1, rowkey, jo))
              })
            case _ =>
          }
          //ES 暂时未实现
       /*   meta.es match {
            case Some(ei) =>
              val ElasticSearchInfo(index, doc_type) = ei
              data.foreach(t => {
                val (rowkey, jo) = t
                ctx.output(OutputTags.ELASTICSEARCH_SINK_TAG, (index, doc_type, rowkey, jo))
              })
            case _ =>
          }*/
          meta.holo match {
            case Some(hi) =>
              val HologresInfo(h0, h1) = hi
              data.foreach(t => {
                val (rowkey, jo) = t
                ctx.output(OutputTags.HOLO_SINK_TAG, (h0, h1, rowkey, jo))
              })
            case _ =>
          }
        }
      }
    }).name("afterOutput")
    val hbaseSource = afterOutput.getSideOutput(OutputTags.HBASE_SINK_TAG)
    hbaseSource
      .map(t => {
        val (table, cf, row, value) = t
        val rowb = row.getBytes("utf-8")
        val cfb = cf.getBytes("utf-8")

        value match {
          case JObject(x) =>
            val put: Put = new Put(rowb)
            x.foreach(t => {
              val (k, v) = t
              val stringv = v match {
                case JNull => ""
                case JNothing => ""
                case x: JObject => x.toJson()
                case JString(x) => x
                case x: JValue => x.toString
              }
              put.addColumn(cfb, k.getBytes("utf-8"), stringv.getBytes("utf-8"))
            })
            val sp = new SealedPut(put)
            (table, sp)
          case _ => throw new RuntimeException("流数据错误")
        }
      }).name("hbaseMap")
      .keyBy(t => {
        val row = new String(t._2.getRowKey, StandardCharsets.UTF_8)
        s"${t._1}${StringUtils.left(row, 2)}"
      })
      .window(TumblingProcessingTimeWindows.of(Time.seconds(10)))
      .trigger(
        new ProcessingTimeCountTrigger[TimeWindow](1000)
      ).aggregate(new AggregateFunction[(String, SealedPut), PutCollection, PutCollection] {
      override def createAccumulator(): PutCollection = {
        new PutCollection
      }

      override def add(value: (String, SealedPut), accumulator: PutCollection): PutCollection = {
        accumulator.put(value._1, value._2)
        accumulator
      }

      override def getResult(accumulator: PutCollection): PutCollection = {
        accumulator
      }

      override def merge(a: PutCollection, b: PutCollection): PutCollection = {
        PutCollection.merge(a, b)
      }
    })

      .addSink(new HbaseSinkFunction).name("HbaseSinkFunction")

//    val holoSource: DataStream[(String, String, String, JObject)] = afterOutput.getSideOutput(OutputTags.HOLO_SINK_TAG)
//    holoSource.map(t => {
//      val (table, pb, row, value) = t
//
//      value match {
//        case JObject(x) =>
//          val x1: immutable.Seq[(String, JValue)] = x
//          x.foreach(t => {
//            val (k, v) = t
//            val stringv = v match {
//              case JNull => ""
//              case JNothing => ""
//              case x: JObject => x.toJson()
//              case JString(x) => x
//              case x: JValue => x.toString
//            }
//          })
//          (table, sp)
//        case _ => throw new RuntimeException("流数据错误")
//      }
//    })

    env.execute("flink start")
  }

}
