package com.winhc.bigdata.flink.streaming.trigger


import org.apache.flink.api.common.functions.ReduceFunction
import org.apache.flink.api.common.state.{ReducingState, ReducingStateDescriptor}
import org.apache.flink.api.common.typeutils.base.LongSerializer
import org.apache.flink.streaming.api.windowing.triggers.Trigger.TriggerContext
import org.apache.flink.streaming.api.windowing.triggers._
import org.apache.flink.streaming.api.windowing.windows.TimeWindow


/**
 * 
 * @author ZhangJi
 * @since 2021-10-11 15:37
 */

object TimeCountTrigger {
  def processingTime[W<:TimeWindow](maxCount: Long): TimeCountTrigger[W] = {
    new ProcessingTimeCountTrigger[W](maxCount)
  }

  def eventTime[W<:TimeWindow](maxCount: Long): TimeCountTrigger[W] = {
    new EventTimeCountTrigger[W](maxCount)
  }
}

abstract class TimeCountTrigger[W <: TimeWindow](maxCount: Long) extends Trigger[Object, W] {


  /**
   * 计数状态
   */
  private val countState: ReducingStateDescriptor[java.lang.Long] = new ReducingStateDescriptor[java.lang.Long](
    "count", new Sum(), LongSerializer.INSTANCE
  )

  /**
   * 这里绝逼是Flink里的坑，在返回`TriggerResult.FIRE_AND_PURGE`的时候不会清空计数
   *
   * @param window 窗信息
   * @param ctx    上下文
   * @return
   */
  protected def fireAndPurge(window: W, ctx: TriggerContext): TriggerResult = {
    clear(window, ctx)
    TriggerResult.FIRE_AND_PURGE
  }

  override def onElement(element: Object, timestamp: Long, window: W, ctx: TriggerContext): TriggerResult = {
    val count: ReducingState[java.lang.Long] = ctx.getPartitionedState(countState)
    count.add(1L)
    if (count.get >= maxCount || timestamp >= window.getEnd) {
      fireAndPurge(window, ctx)
    } else {
      TriggerResult.CONTINUE
    }
  }

  override def clear(window: W, ctx: TriggerContext): Unit = {
    val count: ReducingState[java.lang.Long] = ctx.getPartitionedState(countState)
    count.clear()
    count.add(0L)
  }

  /**
   * 计数求和
   */
  class Sum extends ReduceFunction[java.lang.Long] {
    def reduce(value1: java.lang.Long, value2: java.lang.Long): java.lang.Long = value1 + value2
  }
}


class ProcessingTimeCountTrigger[W <: TimeWindow](maxCount: Long) extends TimeCountTrigger[W](maxCount) {
  override def onProcessingTime(time: Long, window: W, ctx: TriggerContext): TriggerResult = {
    if (time >= window.getEnd) {
      TriggerResult.CONTINUE
    } else {
      fireAndPurge(window, ctx)
    }
  }

  override def onEventTime(time: Long, window: W, ctx: TriggerContext): TriggerResult = {
    TriggerResult.CONTINUE
  }
}

class EventTimeCountTrigger[W <: TimeWindow](maxCount: Long) extends TimeCountTrigger[W](maxCount) {
  override def onProcessingTime(time: Long, window: W, ctx: TriggerContext): TriggerResult = {
    TriggerResult.CONTINUE
  }

  override def onEventTime(time: Long, window: W, ctx: TriggerContext): TriggerResult = {
    if (time >= window.getEnd) {
      TriggerResult.CONTINUE
    } else {
      fireAndPurge(window, ctx)
    }
  }
}